#!/usr/bin/env python3
"""NCBI functions."""

import gzip
import re
from collections import Counter

import ujson
from Bio import SeqIO
from tolkein import tofetch
from tolkein import tofile
from tolkein import tolog
from tqdm import tqdm

LOGGER = tolog.logger(__name__)

GBIF_API = "https://www.gbif.org/api"


def fetch_gbif_taxlist(params):
    """Fetch gbif taxon entries."""
    limit = params.get("limit", 5)
    offset = params.get("offset", 0)
    highertaxon_key = params.get("root", None)
    url = (
        "%s/species/search?advanced=false&rank=SPECIES&rank=SUBSPECIES&status=ACCEPTED&status=DOUBTFUL"
        % GBIF_API
    )
    if highertaxon_key is not None:
        url += "&highertaxon_key=%s" % highertaxon_key
    url += "&limit=%d&offset=%d" % (limit, offset)
    page = tofetch.fetch_url(url)
    data = ujson.decode(page)
    return data


def stream_gbif_taxa(root=None):
    """Stream gbif taxon entries."""
    params = {"root": root, "limit": 50, "offset": 0}
    data = {"results": [], "endOfRecords": False}
    while not data["endOfRecords"]:
        data = fetch_gbif_taxlist(params)
        for entry in data.get("results", []):
            yield (entry["key"], entry)
        params["offset"] += params["limit"]


def prepare_xref_rows(identifiers, meta):
    """Convert identifiers to a set of rows for output."""
    rows = []
    common = {"species": meta["species"]}
    if "family" in meta:
        common.update({"family": meta["family"]})
    if "subspecies" in meta:
        common.update({"subspecies": meta["subspecies"]})
    if "NCBI" in identifiers:
        common.update({"ncbiTaxonId": identifiers["NCBI"]["id"]})
    else:
        common.update({"gbifTaxonId": "GBIF:%s" % identifiers["GBIF"]["id"]})
    for db, entry in identifiers.items():
        row = {**common}
        row.update({"xref": "%s:%s" % (db, str(entry["id"]))})
        row.update({"sourceUrl": entry["url"]})
        row.update({"source": entry["source"]})
        row.update({"sourceSlug": entry["id"]})
        row.update({"sourceStub": entry["url"].replace(str(entry["id"]), "")})
        rows.append(row)
    return rows


def fetch_gbif_identifiers(taxon, *, xrefs=None):
    """Fetch gbif identifiers for a taxon."""
    if xrefs is None:
        xrefs = []
    xrefs = set(["NCBI"] + xrefs)
    url = "%s/wikidata/species/%s?locale=en" % (GBIF_API, str(taxon))
    page = tofetch.fetch_url(url)
    identifiers = {
        "GBIF": {
            "id": taxon,
            "url": "https://www.gbif.org/species/%s" % taxon,
            "source": "GBIF taxonKey",
        }
    }
    if page is not None:
        data = ujson.decode(page)
        for entry in data.get("identifiers", []):
            label = entry.get("label", {}).get("value", "None")
            db = label.split(" ")[0]
            if db in xrefs:
                identifiers.update(
                    {db: {"id": entry["id"], "url": entry["url"], "source": label}}
                )
    return identifiers


def gbif_parser(_params, opts):
    """Parse GBIF taxa and identifiers."""
    parsed = []
    for root in opts["gbif-root"]:
        for key, meta in stream_gbif_taxa(root):
            identifiers = fetch_gbif_identifiers(key, xrefs=opts["gbif-xref"])
            parsed += prepare_xref_rows(identifiers, meta)
    return parsed



#!/usr/bin/env python3

"""
Fill attribute values.

Usage:
    genomehubs fill [--hub-name STRING] [--hub-path PATH] [--hub-version PATH]
                    [--config-file PATH...] [--config-save PATH]
                    [--es-host URL...]
                    [--traverse-infer-ancestors] [--traverse-infer-descendants]
                    [--traverse-infer-both] [--traverse-threads INT]
                    [--traverse-depth INT] [--traverse-root STRING]
                    [--traverse-weight STRING]
                    [-h|--help] [-v|--version]

Options:
    --hub-name STRING             GenomeHubs instance name.
    --hub-path PATH               GenomeHubs instance root directory path.
    --hub-version STR             GenomeHubs instance version string.
    --config-file PATH            Path to YAML file containing configuration options.
    --config-save PATH            Path to write configuration options to YAML file.
    --es-host URL                 ElasticSearch hostname/URL and port.
    --traverse-depth INT          Maximum depth for tree traversal relative to root taxon.
    --traverse-infer-ancestors    Flag to enable tree traversal from tips to root.
    --traverse-infer-descendants  Flag to enable tree traversal from root to tips.
    --traverse-infer-both         Flag to enable tree traversal from tips to root and
                                  back to tips.
    --traverse-root ID            Root taxon id for tree traversal.
    --traverse-threads INT        Number of threads to use for tree traversal. [Default: 1]
    --traverse-weight STRING      Weighting scheme for setting values during tree
                                  traversal.
    -h, --help                    Show this
    -v, --version                 Show version number

Examples:
    # 1. Traverse tree up to taxon_id 7088
    ./genomehubs fill --traverse-root 7088
"""

import platform
import sys
import time
from collections import defaultdict
from multiprocessing import Pool
from statistics import mean
from statistics import median
from statistics import median_high
from statistics import median_low
from statistics import mode
from subprocess import call

from docopt import docopt
from tolkein import tolog
from tqdm import tqdm

from ..lib import hub
from ..lib import taxon
from .attributes import fetch_types
from .config import config
from .es_functions import index_stream
from .es_functions import launch_es
from .es_functions import stream_template_search_results
from .version import __version__

# if platform.system() != "Linux":
#     from multiprocessing import set_start_method

#     set_start_method("fork")

LOGGER = tolog.logger(__name__)


def get_max_depth(es, *, index):
    """Find max depth of root lineage."""
    body = {
        "id": "max_nested_value",
        "params": {"path": "lineage", "field": "node_depth"},
    }
    res = es.search_template(index=index, body=body)
    max_depth = res["aggregations"]["depths"]["max_depth"]["value"]
    return max_depth


def get_max_depth_by_lineage(es, *, index, root):
    """Find max depth of specified root lineage."""
    body = {
        "id": "max_nested_value_by_key_value",
        "params": {
            "path": "lineage",
            "key": "taxon_id",
            "value": root,
            "field": "node_depth",
        },
    }
    res = es.search_template(index=index, body=body)
    max_depth = res["aggregations"]["depths"]["root"]["max_depth"]["value"]
    return max_depth


def stream_nodes_by_root_depth(es, *, index, root, depth, size=10):
    """Get entries by depth of root taxon."""
    if depth > 0:
        body = {
            "id": "taxon_attributes_by_root_depth",
            "params": {"taxon_id": root, "depth": depth},
        }
        return stream_template_search_results(es, index=index, body=body, size=size)
    body = {
        "id": "taxon_attributes_by_taxon_id",
        "params": {"taxon_id": root},
    }
    return stream_template_search_results(es, index=index, body=body)


def stream_descendant_nodes_missing_attributes(es, *, index, attributes, root, size=10):
    """Get entries descended from root that lack one or more attributes."""
    id_list = set()
    for attribute in attributes:
        body = {
            "id": "taxon_missing_attribute_by_ancestor_id",
            "params": {"taxon_id": root, "attribute": attribute["key"]},
        }
        for result in stream_template_search_results(
            es, index=index, body=body, size=size
        ):
            if result["_id"] not in id_list:
                yield result


def apply_summary(summary, values, *, max_value=None, min_value=None):
    """Apply summary statistic functions."""
    summaries = {
        "count": len,
        "max": max,
        "min": min,
        "mean": mean,
        "median": median,
        "median_high": median_high,
        "median_low": median_low,
        "mode": mode,
        "most_common": mode,
        "list": list,
    }
    flattened = []
    for v in values:
        if isinstance(v, list):
            flattened += v
        else:
            flattened.append(v)
    value = summaries[summary](flattened)
    if summary == "max":
        if max_value is not None:
            value = max(value, max_value)
        max_value = value
    if summary == "min":
        if min_value is not None:
            value = min(value, min_value)
        min_value = value
    return value, max_value, min_value


def summarise_attribute_values(
    attribute, meta, *, values=None, max_value=None, min_value=None
):
    """Calculate a single summary value for an attribute."""
    if values is None and "values" not in attribute:
        return None, None, None
    if "summary" in meta:
        value_type = "%s_value" % meta["type"]
        if "values" in attribute:
            if values is None:
                values = [value[value_type] for value in attribute["values"]]
            else:
                values += [value[value_type] for value in attribute["values"]]
        if not values:
            return None, None, None
        idx = 0
        traverse = meta.get("traverse", False)
        traverse_value = None
        if not isinstance(meta["summary"], list):
            meta["summary"] = [meta["summary"]]
        for summary in meta["summary"]:
            value, max_value, min_value = apply_summary(
                summary, values, max_value=max_value, min_value=min_value
            )
            if idx == 0:
                attribute[value_type] = value
                attribute["count"] = len(values)
                attribute["aggregation_method"] = summary
                attribute["aggregation_source"] = "direct"
                traverse_value = value
            elif traverse and summary == traverse:
                traverse_value = value
            if summary != "list":
                if summary.startswith("median"):
                    summary = "median"
            else:
                traverse_value = list(set(traverse_value))
            idx += 1
        return traverse_value, max_value, min_value
    return None, None, None


def summarise_attributes(*, attributes, attrs, meta, parent, parents):
    """Set attribute summary values."""
    changed = False
    for node_attribute in attributes:
        if node_attribute["key"] in attrs:
            summary_value, max_value, min_value = summarise_attribute_values(
                node_attribute, meta[node_attribute["key"]]
            )
            if summary_value is not None:
                changed = True
                if parent is not None:
                    if isinstance(summary_value, list):
                        parents[parent][node_attribute["key"]][
                            "values"
                        ] += summary_value
                    else:
                        parents[parent][node_attribute["key"]]["values"].append(
                            summary_value
                        )
                    if max_value is not None:
                        parents[parent][node_attribute["key"]]["max"] = max(
                            parents[parent][node_attribute["key"]]["max"], max_value
                        )
                    if min_value is not None:
                        parents[parent][node_attribute["key"]]["min"] = min(
                            parents[parent][node_attribute["key"]]["min"], min_value
                        )
    return changed


def set_values_from_descendants(
    *, attributes, descendant_values, meta, parent, taxon_rank, parents
):
    """Set attribute summary values from descendant values."""
    changed = False
    for key, obj in descendant_values.items():
        traverseable = meta[key].get("traverse", False)
        if (
            traverseable
            and "traverse_direction" in meta[key]
            and meta[key]["traverse_direction"] == "down"
        ):
            traverseable = False
        if not traverseable:
            continue
        traverse_limit = meta[key].get("traverse_limit", None)
        if traverse_limit and taxon_rank == traverse_limit:
            continue
        try:
            attribute = next(entry for entry in attributes if entry["key"] == key)
        except StopIteration:
            attribute = {"key": key}
            attributes.append(attribute)
        summary_value, max_value, min_value = summarise_attribute_values(
            attribute,
            meta[key],
            values=obj["values"],
            max_value=obj["max"],
            min_value=obj["min"],
        )
        if summary_value is not None:
            attribute["aggregation_source"] = "descendant"
            changed = True
            if traverse_limit and taxon_rank == traverse_limit:
                continue
            if parent is not None:
                if isinstance(summary_value, list):
                    parents[parent][key]["values"] = list(
                        set(parents[parent][key]["values"] + summary_value)
                    )
                else:
                    parents[parent][key]["values"].append(summary_value)
                if max_value is not None:
                    parents[parent][key]["max"] = max(
                        parents[parent][key]["max"], max_value
                    )
                if min_value is not None:
                    parents[parent][key]["min"] = min(
                        parents[parent][key]["min"], min_value
                    )
    return changed


def traverse_from_tips(es, opts, *, template, root=None, max_depth=None):
    """Traverse a tree, filling in values."""
    if root is None:
        root = opts["traverse-root"]
    if max_depth is None:
        max_depth = get_max_depth_by_lineage(
            es, index=template["index_name"], root=root,
        )
    root_depth = max_depth
    meta = template["types"]["attributes"]
    attrs = set(meta.keys())
    parents = defaultdict(
        lambda: defaultdict(
            lambda: {"max": float("-inf"), "min": float("inf"), "values": []}
        )
    )
    while root_depth >= 0:
        nodes = stream_nodes_by_root_depth(
            es, index=template["index_name"], root=root, depth=root_depth, size=50,
        )
        ctr = 0
        for node in nodes:
            ctr += 1
            changed = False
            if "attributes" in node["_source"] and node["_source"]["attributes"]:
                changed = summarise_attributes(
                    attributes=node["_source"]["attributes"],
                    attrs=attrs,
                    meta=meta,
                    parent=node["_source"].get("parent", None),
                    parents=parents,
                )
            if node["_source"]["taxon_id"] in parents:
                if "attributes" not in node["_source"]:
                    node["_source"]["attributes"] = []
                modified = set_values_from_descendants(
                    attributes=node["_source"]["attributes"],
                    descendant_values=parents[node["_source"]["taxon_id"]],
                    meta=meta,
                    parent=node["_source"].get("parent", None),
                    parents=parents,
                    taxon_rank=node["_source"]["taxon_rank"],
                )
                if not changed:
                    changed = modified
            if changed:
                yield node["_id"], node["_source"]
        root_depth -= 1


def copy_attribute_summary(source, meta):
    """Copy an attribute summary, removing values."""
    dest = {}
    for key in meta["summary"]:
        if key.startswith("median") and "median" in source:
            dest["median"] = source["median"]
        elif key != "list" and key in source:
            dest[key] = source[key]
    try:
        dest["%s_value" % meta["type"]] = source["%s_value" % meta["type"]]
    except KeyError as err:
        print(source)
        print(meta)
        raise (err)
    dest["count"] = source["count"]
    dest["key"] = source["key"]
    return dest


def stream_missing_attributes_at_level(es, *, nodes, attrs, template, level=1):
    """Stream all descendant nodes with missing attributes."""
    for node in nodes:
        taxon_id = node["_source"]["taxon_id"]
        fill_attrs = []
        if "attributes" in node["_source"]:
            for attribute in node["_source"]["attributes"]:
                if attribute["key"] in attrs:
                    fill_attrs.append(attribute)
        if not fill_attrs:
            continue
        meta = template["types"]["attributes"]
        desc_nodes = stream_descendant_nodes_missing_attributes(
            es,
            index=template["index_name"],
            attributes=fill_attrs,
            root=taxon_id,
            size=50,
        )
        anc_attributes = {attr["key"]: attr for attr in fill_attrs}
        for desc_node in desc_nodes:
            skip_attrs = set()
            if "attributes" in desc_node["_source"]:
                for desc_attribute in desc_node["_source"]["attributes"]:
                    if desc_attribute["key"] in anc_attributes:
                        skip_attrs.add(desc_attribute["key"])
            else:
                desc_node["_source"]["attributes"] = []
            for key, anc_attribute in anc_attributes.items():
                if key not in skip_attrs:
                    desc_attribute = copy_attribute_summary(anc_attribute, meta[key])
                    desc_attribute["aggregation_method"] = meta[key]["traverse"]
                    desc_attribute["aggregation_source"] = "ancestor"
                    desc_node["_source"]["attributes"].append(desc_attribute)
            yield desc_node["_id"], desc_node["_source"]


def traverse_from_root(es, opts, *, template, root=None, max_depth=None, log=True):
    """Traverse a tree, filling in values."""
    if root is None:
        root = opts["traverse-root"]
    if max_depth is None:
        max_depth = get_max_depth_by_lineage(
            es, index=template["index_name"], root=root,
        )
    root_depth = max_depth - 1
    meta = template["types"]["attributes"]
    attrs = set({})
    for key, value in meta.items():
        if "traverse" in value and value["traverse"]:
            if "traverse_direction" not in value or value["traverse_direction"] != "up":
                attrs.add(key)
    while root_depth >= 0:
        if log:
            LOGGER.info("Filling values at root depth %d" % root_depth)
        nodes = stream_nodes_by_root_depth(
            es, index=template["index_name"], root=root, depth=root_depth, size=50,
        )
        desc_nodes = stream_missing_attributes_at_level(
            es, nodes=nodes, attrs=attrs, template=template
        )
        index_stream(es, template["index_name"], desc_nodes, _op_type="update", log=log)
        root_depth -= 1


def traverse_tree(es, opts, template, root, max_depth):
    """Propagate values by tree traversal."""
    log = True
    if es is None:
        log = False
        es = launch_es(opts, log=log)
    if "traverse-infer-ancestors" in opts:
        if log:
            LOGGER.info("Inferring ancestral values for root taxon %s", root)
        _success, _failed = index_stream(
            es,
            template["index_name"],
            traverse_from_tips(
                es, opts, template=template, root=root, max_depth=max_depth
            ),
            _op_type="update",
            log=log,
        )
    if "traverse-infer-descendants" in opts:
        if log:
            LOGGER.info("Inferring descendant values for root taxon %s", root)
        traverse_from_root(
            es, opts, template=template, root=root, max_depth=max_depth, log=log
        )


def traverse_helper(params):
    """Wrap traverse_tree for multithreaded traversal."""
    with tolog.DisableLogger():
        traverse_tree(*params)
    return params[3]


def traverse_handler(es, opts, template):
    """Handle single or multi-threaded tree traversal."""
    root = opts["traverse-root"]
    threads = int(opts["traverse-threads"])
    max_depth = get_max_depth_by_lineage(es, index=template["index_name"], root=root)
    subtree_depth = int(opts.get("traverse-depth", 0))
    if threads > 1 and subtree_depth == 0:
        subtree_depth = int(max_depth / 2)
    if threads == 1:
        traverse_tree(es, opts, template, root, max_depth)
        return

    max_depth -= subtree_depth

    nodes = stream_nodes_by_root_depth(
        es, index=template["index_name"], root=root, depth=subtree_depth
    )
    roots = [
        (None, opts, template, node["_source"]["taxon_id"], max_depth) for node in nodes
    ]
    LOGGER.info("Filling values in subtrees")
    with Pool(processes=threads) as p:
        with tqdm(total=len(roots), unit=" subtrees") as pbar:
            for root in p.imap_unordered(traverse_helper, roots):
                pbar.update()
    LOGGER.info("Connecting subtrees")
    traverse_tree(es, opts, template, opts["traverse-root"], subtree_depth)


def main(args):
    """Initialise genomehubs."""
    options = config("fill", **args)
    if "traverse-infer-both" in options["fill"]:
        options["fill"]["traverse-infer-ancestors"] = True
        options["fill"]["traverse-infer-descendants"] = True

    # Start Elasticsearch
    es = launch_es(options["fill"])

    # Post search scripts
    hub.post_search_scripts(es)

    LOGGER.info("Filling values")

    types = fetch_types(es, "taxon", options["fill"])

    if "taxonomy-source" in options["fill"]:
        for taxonomy_name in options["fill"]["taxonomy-source"]:
            template = taxon.index_template(taxonomy_name, options["fill"])
            if types:
                template["types"]["attributes"] = types
            if "traverse-root" in options["fill"]:
                traverse_handler(es, options["fill"], template)


def cli():
    """Entry point."""
    if len(sys.argv) == sys.argv.index(__name__.split(".")[-1]) + 1:
        args = docopt(__doc__, argv=[])
    else:
        args = docopt(__doc__, version=__version__)
    main(args)


if __name__ == "__main__":
    cli()



#!/usr/bin/env python3
"""Hub functions."""

import os
import re
import sys
from copy import deepcopy
from pathlib import Path

from tolkein import tofile
from tolkein import tolog

LOGGER = tolog.logger(__name__)
MIN_INTEGER = -(2 ** 31)
MAX_INTEGER = 2 ** 31 - 1


def setup(opts):
    """Set up directory for this GenomeHubs instance and handle reset."""
    Path(opts["hub-path"]).mkdir(parents=True, exist_ok=True)
    return True


def load_types(name, *, part="types"):
    """Read a types file from the templates directory."""
    script_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    types = None
    try:
        types_file = os.path.join(script_dir, "templates", "%s.%s.yaml" % (name, part))
        types = tofile.load_yaml(types_file)
        try:
            for key, value in types.items():
                try:
                    enum = set(value["constraint"]["enum"])
                    value["constraint"]["enum"] = enum
                except KeyError:
                    pass
        except AttributeError:
            pass
    except Exception:
        pass
    if types and "file" in types and "attributes" in types:
        for key, value in types["file"].items():
            for attr in types["attributes"].values():
                if key not in ("format", "header", "name") and key not in attr:
                    attr[key] = value
    return types


def index_templator(parts, opts):
    """Index template helper function."""
    script_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    mapping_file = os.path.join(script_dir, "templates", parts[0] + ".json")
    types = load_types(parts[0])
    names = load_types(parts[0], part="names")
    template = {
        "name": parts[0],
        "index_name": opts["hub-separator"].join(parts),
        "mapping": tofile.load_yaml(mapping_file),
        "types": types,
        "names": names,
    }
    return template


def order_parsed_fields(parsed, types, names=None):
    """Order parsed fields using a template file."""
    columns = {}
    fields = {}
    ctr = 0
    types = deepcopy(types)
    if names is not None:
        for group, entries in names.items():
            if group not in types:
                types[group] = deepcopy(entries)
            else:
                for field, attrs in entries.items():
                    if isinstance(attrs, dict):
                        if field not in types[group]:
                            types[group][field] = deepcopy(attrs)
                        elif types[group][field]["header"] != attrs["header"]:
                            types[group]["names_%s" % field] = deepcopy(attrs)
    for group, entries in types.items():
        for field, attrs in entries.items():
            header = False
            try:
                for key, value in attrs.items():
                    if key == "index":
                        if value not in columns:
                            columns.update({value: field})
                            fields.update({field: value})
                    elif key == "header":
                        header = value
                if header and header not in fields:
                    columns.update({ctr: header})
                    fields.update({header: ctr})
                    ctr += 1
            except AttributeError:
                pass
    order = [x[0] for x in sorted(fields.items(), key=lambda x: x[1])]
    data = [order]
    for entry in parsed:
        row = [entry.get(field, "None") for field in order]
        data.append(row)
    return data


def post_search_scripts(es):
    """POST ElasticSearch search scripts."""
    script_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    search_script_dir = os.path.join(script_dir, "templates", "scripts")
    directory = os.fsencode(search_script_dir)
    for script_file in os.listdir(directory):
        filename = os.fsdecode(script_file)
        if filename.endswith(".json"):
            body = tofile.load_yaml(os.path.join(search_script_dir, filename))
            script_id = filename.replace(".json", "")
            with tolog.DisableLogger():
                es.put_script(id=script_id, body=body)


def byte_type_constraint(value):
    """Test byte_type constraint."""
    if value >= -128 and value <= 128:
        return True
    return False


def integer_type_constraint(value):
    """Test integer_type constraint."""
    if value >= MIN_INTEGER and value <= MAX_INTEGER:
        return True
    return False


def short_type_constraint(value):
    """Test short_type constraint."""
    if value >= -32768 and value <= 32767:
        return True
    return False


def min_value_constraint(value, limit):
    """Test minimum value constraint."""
    if value >= limit:
        return True
    return False


def max_value_constraint(value, limit):
    """Test maximum value constraint."""
    if value <= limit:
        return True
    return False


def enum_constraint(value, enum):
    """Test value in predefined set constraint."""
    if value in enum:
        return True
    return False


def test_constraint(value, constraint):
    """Test value against constraint."""
    constraints = {
        "byte": byte_type_constraint,
        "integer": integer_type_constraint,
        "short": short_type_constraint,
        "min": min_value_constraint,
        "max": max_value_constraint,
        "enum": enum_constraint,
    }
    for key in constraint:
        if key in constraints:
            if not constraints[key](value, constraint[key]):
                return False
    return True


def convert_lat_lon(location):
    """Convert lat and lon to array notation."""
    if isinstance(location, list):
        return ",".join(location)
    if location.endswith(("E", "W")):
        sign = {"N": "", "E": "", "S": "-", "W": "-"}
        parts = re.split(r"\s*([NESW])\s*", location)
        string = "%s%s,%s%s" % (sign[parts[1]], parts[0], sign[parts[3]], parts[2])
        return string
    if location:
        return None
    return ""


def convert_to_type(key, value, to_type):
    """Convert values to type."""
    if to_type in {"byte", "integer", "long", "short"}:
        try:
            value = int(value)
        except ValueError:
            value = None
    elif to_type in {
        "double",
        "float",
        "half_float",
        "1dp",
        "2dp",
        "3dp",
        "4dp",
    }:
        try:
            value = float(value)
        except ValueError:
            value = None
    elif to_type == "geo_point":
        value = convert_lat_lon(value)
    else:
        value = str(value)
    # if value is None:
    #     LOGGER.warning(
    #         "%s value %s is not a valid %s", key, str(value), to_type,
    #     )
    return value


def calculator(value, operation):
    """Template-based calculator."""
    value1, operator, value2 = operation.split(" ")
    if value1 == "{}":
        value1 = float(value)
        value2 = float(value2)
    elif value2 == "{}":
        value2 = float(value)
        value1 = float(value1)
    if operator == "+":
        return value1 + value2
    elif operator == "-":
        return value1 - value2
    elif operator == "*":
        return value1 * value2
    elif operator == "/":
        try:
            return value1 / value2
        except ZeroDivisionError:
            return None
    elif operator == "%":
        if operator == "%":
            return value1 % value2
    else:
        return None


def validate_values(values, key, types):
    """Validate values."""
    validated = []
    key_type = types[key]["type"]
    for value in values:
        if "function" in types[key]:
            try:
                value = calculator(value, types[key]["function"])
            except ValueError:
                continue
        value = convert_to_type(key, value, key_type)
        if value is None:
            continue
        if isinstance(value, str):
            if not value:
                continue
            try:
                value = types[key]["translate"][value]
            except KeyError:
                pass
        try:
            valid = test_constraint(value, key_type)
            if valid:
                valid = test_constraint(value, types[key]["constraint"])
        except KeyError:
            valid = True
        if valid:
            validated.append(value)
        elif value:
            LOGGER.warning("%s is not a valid %s value", str(value), key)
    return validated


def apply_value_template(prop, value, attribute, *, taxon_types, has_taxon_data):
    """Set value using template."""
    template = re.compile(r"^(.*?\{\{)(.+)(\}\}.*)$")
    new_prop = prop.replace("taxon_", "")
    match = template.match(str(value))
    if match:
        groups = match.groups()
        if groups[1] and groups[1] in attribute:
            has_taxon_data = True
            new_value = attribute[groups[1]]
            if groups[0] != "{{":
                new_value = "%s%s" % (groups[0].replace("{{", ""), str(new_value),)
            if groups[2] != "}}":
                new_value = "%s%s" % (str(new_value), groups[2].replace("}}", ""),)
            taxon_types.update({new_prop: new_value})
            taxon_types.update({"group": "taxon"})
            if prop in taxon_types:
                del taxon_types[prop]
    else:
        has_taxon_data = True
        taxon_types.update({new_prop: value})
        taxon_types.update({"group": "taxon"})
        if prop in taxon_types:
            del taxon_types[prop]
    return has_taxon_data


def add_attributes(
    entry, types, *, attributes=None, source=None, attr_type="attributes", meta=None
):
    """Add attributes to a document."""
    if attributes is None:
        attributes = []
    taxon_attributes = []
    taxon_types = {}
    if meta is None:
        meta = {}
    attribute_values = {}
    for key, values in entry.items():
        if key in types:
            if not isinstance(values, list):
                values = [values]
            validated = validate_values(values, key, types)
            # TODO: handle invalid values
            if validated:
                if len(validated) == 1:
                    validated = validated[0]
                if attr_type == "attributes":
                    attribute = {"key": key, "%s_value" % types[key]["type"]: validated}
                    attribute_values.update({key: attribute})
                else:
                    attribute = {"identifier": validated, "class": key}
                attribute.update(meta)
                if source is not None:
                    attribute.update({"source": source})
                attributes.append(attribute)
    if attribute_values:
        for attribute in attributes:
            has_taxon_data = False
            taxon_attribute = {**attribute}
            taxon_attribute_types = {**types[attribute["key"]]}
            # taxon_props = []
            for prop, value in types[attribute["key"]].items():
                if prop.startswith("taxon_"):
                    has_taxon_data = apply_value_template(
                        prop,
                        value,
                        attribute,
                        taxon_types=taxon_attribute_types,
                        has_taxon_data=has_taxon_data,
                    )
                    # taxon_props.append(prop)
            # for prop in taxon_props:
            #     del types[attribute["key"]][prop]
            if has_taxon_data:
                taxon_attribute.update({"key": taxon_attribute_types["name"]})
                # taxon_attribute.update({"name": taxon_attribute_types["key"]})
                taxon_attributes.append(taxon_attribute)
                taxon_types.update(
                    {taxon_attribute_types["name"]: taxon_attribute_types}
                )
    return attributes, taxon_attributes, taxon_types


def chunks(arr, n):
    """Yield successive n-sized chunks from arr."""
    for i in range(0, len(arr), n):
        yield arr[i : i + n]


def add_attribute_values(existing, new, *, raw=True):
    """Add attribute values to records."""
    indices = {}
    for index, entry in enumerate(existing):
        indices[entry["key"]] = index
    index = len(existing)
    for group in new:
        if not isinstance(group, list):
            group = [group]
        for entry in group:
            if raw:
                arr = []
                if indices and entry["key"] in indices:
                    arr = existing[indices[entry["key"]]]["values"]
                else:
                    existing.append({"key": entry["key"], "values": arr})
                    indices[entry["key"]] = index
                    index += 1
                del entry["key"]
                arr.append(entry)
            else:
                existing.append(
                    {
                        **entry,
                        "count": 1,
                        "aggregation_method": "unique",
                        "aggregation_source": "direct",
                    }
                )


def validate_types_file(types_file, dir_path):
    """Validate types file."""
    try:
        types = tofile.load_yaml(str(types_file.resolve()))
    except Exception:
        LOGGER.error("Unable to open types file %s", str(types_file.resolve()))
        sys.exit(1)
    if "taxonomy" not in types:
        LOGGER.error("Types file contains no taxonomy information")
    if "file" not in types or "name" not in types["file"]:
        LOGGER.error("No data file name in types file")
    defaults = {"attributes": {}, "metadata": {}}
    for key, value in types["file"].items():
        if key.startswith("display") or key.startswith("taxon"):
            defaults["attributes"].update({key: value})
        elif key.startswith("source"):
            defaults["metadata"].update({key: value})
    types.update({"defaults": defaults})
    data = tofile.open_file_handle(Path(dir_path) / types["file"]["name"])
    return types, data


def process_row(types, row):
    """Process a row of data."""
    data = {
        "attributes": {},
        "identifiers": {},
        "metadata": {},
        "taxon_names": {},
        "taxonomy": {},
        "taxon_attributes": {},
    }
    for key in types["defaults"].keys():
        if key in types:
            for entry in types[key].values():
                entry = {
                    **types["defaults"][key],
                    **entry,
                }
        elif key == "metadata":
            data["metadata"] = {**types["defaults"]["metadata"]}
    for group in data.keys():
        if group in types:
            for key, meta in types[group].items():
                try:
                    if isinstance(meta["index"], list):
                        char = meta.get("join", "")
                        values = [row[i] for i in meta["index"]]
                        if all(values):
                            value = char.join(values)
                        else:
                            continue
                    else:
                        value = row[meta["index"]]
                    if "separator" in meta and any(
                        sep in value for sep in meta["separator"]
                    ):
                        separator = "|".join(meta["separator"])
                        data[group][key] = re.split(rf"\s*{separator}\s*", value)
                    else:
                        data[group][key] = value
                except Exception as err:
                    LOGGER.warning("Cannot parse row '%s'" % str(row))
                    raise err
    taxon_data = {}
    taxon_types = {}
    for attr_type in list(["attributes", "identifiers"]):
        if data[attr_type]:
            (
                data[attr_type],
                taxon_data[attr_type],
                taxon_types[attr_type],
            ) = add_attributes(
                data[attr_type],
                types[attr_type],
                attr_type=attr_type,
                meta=data["metadata"],
            )
        else:
            data[attr_type] = []
    return data, taxon_data, taxon_types.get("attributes", {})


def set_column_indices(types, header):
    """Use header to set indices for named columns."""
    headers = {title: index for index, title in enumerate(header)}
    for section, entries in types.items():
        for key, value in entries.items():
            if isinstance(value, dict):
                if "header" in value:
                    index = headers.get(value["header"], None)
                    if index is not None:
                        value.update({"index": index})



#!/usr/bin/env python3
"""Wikidata functions."""

import gzip
import re
import sys
import time
from collections import Counter
from datetime import datetime
from datetime import timedelta

import ujson
from Bio import SeqIO
from SPARQLWrapper import JSON
from SPARQLWrapper import SPARQLWrapper
from tolkein import tofetch
from tolkein import tofile
from tolkein import tolog
from tqdm import tqdm

LOGGER = tolog.logger(__name__)

BTK_API = "https://blobtoolkit.genomehubs.org/api/v1"
BTK_VIEW = "https://blobtoolkit.genomehubs.org/view"


def fetch_btk_datasets(params):
    """Fetch wikidata taxon entries."""
    root = params.get("root", "Eukaryota")
    url = "%s/search/%s" % (BTK_API, root)
    page = tofetch.fetch_url(url)
    datasets = ujson.decode(page)
    return datasets


def stream_btk_datasets(root=None):
    """Stream wikidata taxon entries."""
    params = {"root": root}
    datasets = fetch_btk_datasets(params)
    if datasets:
        for dataset in datasets:
            yield dataset["id"], dataset


def extract_btk_stats(meta):
    """Extract BlobToolKit stats to top level."""
    summaryStats = meta.pop("summaryStats", {})
    meta["source"] = "BlobToolKit"
    meta["sourceSlug"] = meta["id"]
    meta["sourceStub"] = "https://blobtoolkit.genomehubs.org/view/dataset/"
    # meta["xref"] = "BTK:%s" % meta["id"]
    if "busco" in summaryStats:
        for busco_lineage, busco_stats in summaryStats["busco"].items():
            meta["busco_lineage"] = busco_lineage
            meta["busco_string"] = busco_stats["string"]
            meta["busco_complete"] = busco_stats["c"] / busco_stats["t"] * 100
            break
    if "stats" in summaryStats:
        meta["nohit"] = summaryStats["stats"]["noHit"] * 100
        meta["target"] = summaryStats["stats"]["target"] * 100
    if "baseComposition" in summaryStats:
        meta["at_percent"] = summaryStats["baseComposition"]["at"] * 100
        meta["gc_percent"] = summaryStats["baseComposition"]["gc"] * 100
        meta["n_percent"] = summaryStats["baseComposition"]["n"] * 100
    if len(meta["taxon_name"]) > len(meta["species"]):
        meta["subspecies"] = meta["taxon_name"]


def describe_btk_files(meta):
    """Generate analysis descriptions and links for BlobToolKit."""
    plots = ["cumulative", "snail"]
    summaryStats = meta.get("summaryStats", {})
    if "readMapping" in summaryStats and summaryStats["readMapping"]:
        plots.append("blob")
    files = []
    for plot in plots:
        if plot == "blob":
            url = "%s/image/%s/%s/circle?format=png" % (BTK_API, meta["id"], plot)
        else:
            url = "%s/image/%s/%s?format=png" % (BTK_API, meta["id"], plot)
        obj = {
            "name": "%s.png" % plot,
            "url": url,
            "link": "%s/%s/dataset/%s/%s" % (BTK_VIEW, meta["id"], meta["id"], plot),
            "analysis_id": "btk-%s" % meta["id"],
            "description": "a %s plot from BlobToolKit analysis %s"
            % (plot, meta["id"]),
            "title": "%s plot %s" % (plot, meta["id"]),
            "command": "blobtoolkit pipeline",
            "assembly_id": meta["accession"],
            "taxon_id": str(meta["taxid"]),
            "analysis": {
                "name": "BlobToolKit",
                "title": "BlobToolKit analysis of %s" % meta["accession"],
                "description": "Analysis of public assembly %s using BlobToolKit"
                % meta["accession"],
            },
        }
        files.append(obj)
    return files


def btk_parser(_params, opts):
    """Parse BlobToolKit assemblies."""
    parsed = []
    analyses = []
    for root in opts["btk-root"]:
        for key, meta in stream_btk_datasets(root):
            files = describe_btk_files(meta)
            analyses += files
            extract_btk_stats(meta)
            # TODO: set xref
            # TODO: add to file/analysis index
            # print(lineage)
            # print(
            #     "%s: %s (%s)" % (key, meta["name"]["value"], meta["rankLabel"]["value"])
            # )
            # identifiers = fetch_gbif_identifiers(key, xrefs=opts["gbif-xref"])
            # parsed += prepare_xref_rows(key, meta, lineage)
            parsed.append(meta)
    return (parsed, analyses)



#!/usr/bin/env python3

"""Taxon methods."""

import os
import sys
from collections import defaultdict

from tolkein import tofile
from tolkein import tolog
from tqdm import tqdm

from .es_functions import EsQueryBuilder
from .es_functions import document_by_id
from .es_functions import index_stream
from .es_functions import query_keyword_value_template
from .es_functions import query_value_template
from .hub import add_attribute_values
from .hub import add_attributes
from .hub import chunks
from .hub import index_templator
from .taxonomy import index_template as taxonomy_index_template

LOGGER = tolog.logger(__name__)


def index_template(taxonomy_name, opts):
    """Index template (includes name, mapping and types)."""
    parts = ["taxon", taxonomy_name, opts["hub-name"], opts["hub-version"]]
    template = index_templator(parts, opts)
    return template


def lookup_taxon_by_taxid(es, taxon_id, taxonomy_template):
    """Lookup taxon in taxonomy index by taxon_id."""
    query = EsQueryBuilder()
    query.es_match("taxon_id", taxon_id)
    with tolog.DisableLogger():
        res = es.search(index=taxonomy_template["index_name"], body=query.write())
    if res["hits"]["total"]["value"] == 1:
        return res["hits"]["hits"][0]["_source"]
    return None


def lookup_taxa_by_taxon_id(es, values, template, *, return_type="list"):
    """Retrieve existing taxa from index."""
    taxa = []
    if return_type == "dict":
        taxa = {}
    res = document_by_id(
        es, ["taxon_id-%s" % value for value in values], template["index_name"]
    )
    if res is not None:
        if return_type == "list":
            taxa = [key.replace("taxon_id-", "") for key in res.keys()]
        else:
            for key, source in res.items():
                taxon_id = key.replace("taxon_id-", "")
                taxa.update({taxon_id: {"_id": key, "_source": source}})
    # taxon_res = query_keyword_value_template(
    #     es,
    #     "attributes_names_by_keyword_value",
    #     "taxon_id",
    #     values,
    #     index=template["index_name"],
    # )
    # if taxon_res is not None:
    #     for response in taxon_res["responses"]:
    #         if "hits" in response and response["hits"]["total"]["value"] == 1:
    #             if return_type == "list":
    #                 taxa.append(response["hits"]["hits"][0])
    #             else:
    #                 taxa[response["hits"]["hits"][0]["_source"]["taxon_id"]] = response[
    #                     "hits"
    #                 ]["hits"][0]
    #         elif return_type == "list":
    #             taxa.append(None)
    return taxa


def lookup_missing_taxon_ids(
    es, without_ids, opts, *, with_ids=None, blanks=set(["NA", "None"])
):
    """Lookup taxon ID based on available taxonomic information."""
    if with_ids is None:
        with_ids = {}
    # TODO: set this list from types file
    ranks = [
        "subspecies",
        "species",
        "genus",
        "family",
        "order",
        "class",
        "subphylum",
        "phylum",
    ]
    found_keys = []
    pbar = tqdm(total=len(without_ids.keys()))
    for key, arr in without_ids.items():
        pbar.update(1)
        for obj in arr:
            if "taxonomy" not in obj:
                continue
            for index, rank in enumerate(ranks):
                if rank not in obj["taxonomy"] or obj["taxonomy"][rank] in blanks:
                    continue
                taxon_ids, name_class = lookup_taxon(
                    es, obj["taxonomy"][rank], opts, rank=rank
                )
                if index == 1 and not taxon_ids:
                    break
                if len(taxon_ids) == 1:
                    if taxon_ids[0] in with_ids:
                        with_ids[taxon_ids[0]].append(obj)
                    else:
                        obj["attributes"] = [obj["attributes"]]
                        with_ids[taxon_ids[0]] = [obj]
                        LOGGER.debug(
                            "Matched %s with taxon_id %s",
                            obj["taxonomy"][rank],
                            taxon_ids[0],
                        )
                    found_keys.append(key)
                break
                # TODO: Allow lookup for higher taxa
                # if not taxon_ids:
                #     continue
                for anc_rank in ranks[(index + 1) :]:
                    if (
                        anc_rank not in obj["taxonomy"]
                        or obj["taxonomy"][anc_rank] in blanks
                    ):
                        continue
                    taxon_ids = lookup_taxon_within_lineage(
                        es,
                        obj["taxonomy"][rank],
                        obj["taxonomy"][anc_rank],
                        opts,
                        rank=rank,
                        anc_rank=anc_rank,
                        return_type="taxon_id",
                        name_class=name_class,
                    )
                    if taxon_ids:
                        if len(taxon_ids) == 1:
                            if taxon_ids[0] in with_ids:
                                with_ids[taxon_ids[0]].append(obj)
                            else:
                                obj["attributes"] = [obj["attributes"]]
                                with_ids[taxon_ids[0]] = [obj]
                                LOGGER.debug(
                                    "Matched %s with taxon_id %s",
                                    obj["taxonomy"][rank],
                                    taxon_ids[0],
                                )
                            found_keys.append(key)
                        else:
                            LOGGER.warn(
                                "Taxon name %s is ambiguous within %s",
                                obj["taxonomy"][rank],
                                obj["taxonomy"][anc_rank],
                            )
                        break
                break
    pbar.close()
    found_ids = {taxon_id: True for taxon_id in with_ids.keys()}
    for key in found_keys:
        without_ids.pop(key, None)
        found_ids[key] = True
    return with_ids, without_ids, found_ids


def fix_missing_ids(
    es,
    opts,
    without_ids,
    *,
    types,
    taxon_template,
    failed_rows,
    with_ids=None,
    blanks=set(["NA", "None"]),
    header=None,
):
    """Find or create taxon IDs for rows without."""
    if with_ids is None:
        with_ids = {}
    if without_ids:
        # TODO: support multiple taxonomies
        LOGGER.info("Looking up %d missing taxon IDs", len(without_ids.keys()))
        with_ids, without_ids, found_ids = lookup_missing_taxon_ids(
            es, without_ids, opts, with_ids=with_ids, blanks=blanks
        )
        # create new taxon IDs
        if "taxonomy" in types and "alt_taxon_id" in types["taxonomy"]:
            LOGGER.info(
                "Using alt_taxon_id to fill in %d missing taxon IDs",
                len(without_ids.keys()),
            )
            created_ids = create_taxa(
                es,
                opts,
                data=without_ids,
                blanks=blanks,
                taxon_template=taxon_template,
            )
            for created_id in created_ids:
                if created_id in without_ids:
                    with_ids[created_id] = without_ids[created_id]
                    found_ids[created_id] = True
                    del without_ids[created_id]
    if without_ids and failed_rows:
        for key, value in found_ids.items():
            if key in failed_rows:
                del failed_rows[key]
        if failed_rows:
            LOGGER.info(
                "Unable to associate %d records with taxon IDs", len(failed_rows)
            )
            data = []
            exception_key = "%s-exception" % opts["index"]
            dir_key = "%s-dir" % opts["index"]
            if exception_key in opts and opts[exception_key]:
                outdir = opts[exception_key]
            else:
                outdir = "%s/exceptions" % opts[dir_key]
            os.makedirs(outdir, exist_ok=True)
            outfile = "%s/%s" % (outdir, types["file"]["name"])
            if header:
                data.append(header)
            for rows in failed_rows.values():
                for row in rows:
                    data.append(row)
            LOGGER.info(
                "Writing %d records to exceptions file '%s", len(data) - 1, outfile
            )
            tofile.write_file(outfile, data)
    return with_ids, without_ids


def add_taxonomy_info_to_meta(meta, source):
    """Add taxonomy information to a metadata dict."""
    keys = {"lineage", "parent", "scientific_name", "taxon_names", "taxon_rank"}
    for key in keys:
        if key in source:
            meta[key] = source[key]


def stream_taxa(taxa):
    """Stream dict of taxa for indexing."""
    for taxon_id, value in taxa.items():
        yield "taxon_id-%s" % taxon_id, value


def get_taxa_to_create(
    es, opts, *, taxonomy_name="ncbi", taxon_ids=None, asm_by_taxon_id=None,
):
    """Create a dict of taxa to create."""
    taxa_to_create = {}
    if not taxon_ids:
        return {}
    if asm_by_taxon_id is None:
        asm_by_taxon_id = {}
    taxonomy_template = taxonomy_index_template(taxonomy_name, opts)
    taxonomy_res = query_value_template(
        es, "taxonomy_node_by_taxon_id", taxon_ids, taxonomy_template["index_name"],
    )
    if taxonomy_res is None:
        LOGGER.error(
            "Could not connect to taxonomy index '%s'", taxonomy_template["index_name"],
        )
        sys.exit(1)
    ancestors = set()
    for taxonomy_result in taxonomy_res["responses"]:
        if taxonomy_result["hits"]["total"]["value"] == 1:
            source = taxonomy_result["hits"]["hits"][0]["_source"]
            taxa_to_create[source["taxon_id"]] = source
            for ancestor in source["lineage"]:
                ancestors.add(ancestor["taxon_id"])
            if source["taxon_id"] in asm_by_taxon_id:
                for asm in asm_by_taxon_id[source["taxon_id"]]:
                    add_taxonomy_info_to_meta(asm, source)
    taxonomy_res = query_value_template(
        es,
        "taxonomy_node_by_taxon_id",
        list(ancestors),
        taxonomy_template["index_name"],
    )
    if taxonomy_res and "responses" in taxonomy_res:
        for taxonomy_result in taxonomy_res["responses"]:
            if taxonomy_result["hits"]["total"]["value"] == 1:
                source = taxonomy_result["hits"]["hits"][0]["_source"]
                taxa_to_create[source["taxon_id"]] = source
    return taxa_to_create


def find_or_create_taxa(es, opts, *, taxon_ids, taxon_template, asm_by_taxon_id=None):
    """Find taxa in taxon index or create new taxon entries from taxonomy."""
    taxa = lookup_taxa_by_taxon_id(
        es, list(taxon_ids), taxon_template, return_type="dict"
    )
    missing_taxa = [taxon_id for taxon_id in taxon_ids if taxon_id not in taxa]
    to_create = get_taxa_to_create(
        es,
        opts,
        taxonomy_name="ncbi",
        taxon_ids=missing_taxa,
        asm_by_taxon_id=asm_by_taxon_id,
    )
    index_stream(
        es, taxon_template["index_name"], stream_taxa(to_create),
    )
    taxa.update(
        {
            taxon_id: {"_id": "taxon_id-%s" % taxon_id, "_source": obj}
            for taxon_id, obj in to_create.items()
        }
    )
    return taxa


def list_ancestors(taxa):
    """Convert a dict of taxa to a list of ancestral taxon_ids."""
    ancestors = set()
    for doc in taxa.values():
        if "_source" in doc:
            lineage = doc["_source"].get("lineage", [])
        else:
            lineage = doc.get("lineage", [])
        for entry in lineage:
            ancestors.add(entry["taxon_id"])
    return list(ancestors)


def add_names_to_list(existing, new, *, blanks=set({"NA", "None"})):
    """Add names to a list if they do not already exist."""
    names = defaultdict(dict)
    for entry in existing:
        names[entry["class"]][entry["name"]] = True
    for name_class, name in new.items():
        name_class = name_class.replace("_", " ")
        if (
            name not in blanks
            and name_class not in names
            and name not in names[name_class]
        ):
            existing.append({"name": name, "class": name_class})
            names[name_class][name] = True


def add_names_and_attributes_to_taxa(
    es, data, opts, *, template, blanks=set(["NA", "None"])
):
    """Add names and attributes to taxa."""
    for values in chunks(list(data.keys()), 500):
        # taxa = lookup_taxa_by_taxon_id(es, values, template, return_type="list")
        all_taxa = find_or_create_taxa(
            es, opts, taxon_ids=values, taxon_template=template,
        )
        taxa = []
        for taxon_id in values:
            if taxon_id in all_taxa:
                taxa.append(all_taxa[taxon_id])
        for doc in taxa:
            if doc is not None:
                taxon_data = data[doc["_source"]["taxon_id"]]
                taxon_names = {}
                attributes = []
                for entry in taxon_data:
                    if "attributes" in entry:
                        attributes = attributes + entry["attributes"]
                    if "taxon_names" in entry:
                        taxon_names.update(entry["taxon_names"])
                if "taxon_names" not in doc["_source"]:
                    doc["_source"]["taxon_names"] = []
                add_names_to_list(
                    doc["_source"]["taxon_names"], taxon_names, blanks=blanks
                )
                if (
                    "attributes" not in doc["_source"]
                    or not doc["_source"]["attributes"]
                ):
                    doc["_source"]["attributes"] = []
                add_attribute_values(doc["_source"]["attributes"], attributes)
                yield doc["_id"], doc["_source"]


def lookup_taxon_within_lineage(
    es,
    name,
    lineage,
    opts,
    *,
    rank=None,
    anc_rank=None,
    return_type="taxon",
    name_class="scientific",
):
    """Lookup taxon ID in a specified lineage."""
    template = index_template(opts["taxonomy-source"][0], opts)
    body = {
        "id": "taxon_by_lineage",
        "params": {
            "taxon": name,
            "rank": rank,
            "lineage": lineage,
            "anc_rank": anc_rank,
        },
    }
    if name_class == "any":
        body.update({"id": "taxon_by_any_name_by_lineage"})
    with tolog.DisableLogger():
        res = es.search_template(
            body=body, index=template["index_name"], rest_total_hits_as_int=True
        )
    if "hits" in res and res["hits"]["total"] > 0:
        if return_type == "taxon_id":
            return [hit["_source"]["taxon_id"] for hit in res["hits"]["hits"]]
        return [hit for hit in res["hits"]["hits"]]
    index = template["index_name"].replace("taxon", "taxonomy")
    with tolog.DisableLogger():
        res = es.search_template(body=body, index=index, rest_total_hits_as_int=True)
    if "hits" in res and res["hits"]["total"] > 0:
        if return_type == "taxon_id":
            return [hit["_source"]["taxon_id"] for hit in res["hits"]["hits"]]
        return [hit for hit in res["hits"]["hits"]]
    return []


def lookup_taxon(
    es, name, opts, *, rank=None, name_class="scientific", return_type="taxon_id"
):
    """Lookup taxon ID."""
    taxa = []
    template = index_template(opts["taxonomy-source"][0], opts)
    body = {
        "id": "taxon_by_name",
        "params": {"taxon": name, "rank": rank},
    }
    if name_class == "any":
        body.update({"id": "taxon_by_any_name"})
    index = template["index_name"]
    with tolog.DisableLogger():
        res = es.search_template(body=body, index=index, rest_total_hits_as_int=True)
    if "hits" in res and res["hits"]["total"] > 0:
        if return_type == "taxon_id":
            taxa = [hit["_source"]["taxon_id"] for hit in res["hits"]["hits"]]
        else:
            taxa = [hit for hit in res["hits"]["hits"]]
    else:
        template = taxonomy_index_template(opts["taxonomy-source"][0], opts)
        index = template["index_name"]
        with tolog.DisableLogger():
            res = es.search_template(
                body=body, index=index, rest_total_hits_as_int=True
            )
        if "hits" in res and res["hits"]["total"] > 0:
            if return_type == "taxon_id":
                taxa = [hit["_source"]["taxon_id"] for hit in res["hits"]["hits"]]
            else:
                taxa = [hit for hit in res["hits"]["hits"]]
    if not taxa and opts["taxon-lookup"] == "any" and name_class != "any":
        taxa, name_class = lookup_taxon(
            es, name, opts, rank=rank, name_class="any", return_type=return_type
        )
    return taxa, name_class


def generate_ancestral_taxon_id(name, rank, *, alt_taxon_id=None, taxon_ids=None):
    """Generate an ancestral taxon ID."""
    if taxon_ids is None:
        taxon_ids = set({})
    increment = 0
    while True:
        # TODO: make robust to imports from separate files
        anc_taxon_id = "anc_%s" % name
        if increment:
            anc_taxon_id += "_%d" % increment
        if anc_taxon_id not in taxon_ids:
            taxon_ids.add(anc_taxon_id)
            return anc_taxon_id
        increment += 1


def create_descendant_taxon(taxon_id, rank, name, closest_taxon):
    """Set taxon and lineage information for a descendant taxon."""
    desc_taxon = {
        "_id": "taxon_id-%s" % taxon_id,
        "_source": {
            "taxon_id": taxon_id,
            "taxon_rank": rank,
            "scientific_name": name,
            "parent": closest_taxon["_source"]["taxon_id"],
            "taxon_names": [{"class": "scientific name", "name": name}],
        },
    }
    lineage = [
        {
            "node_depth": 1,
            "taxon_id": closest_taxon["_source"]["taxon_id"],
            "taxon_rank": closest_taxon["_source"]["taxon_rank"],
            "scientific_name": closest_taxon["_source"]["scientific_name"],
        }
    ]
    if "lineage" in closest_taxon["_source"]:
        for ancestor in closest_taxon["_source"]["lineage"]:
            lineage.append(
                {
                    "taxon_id": ancestor["taxon_id"],
                    "taxon_rank": ancestor["taxon_rank"],
                    "scientific_name": ancestor["scientific_name"],
                    "node_depth": ancestor["node_depth"] + 1,
                }
            )
    desc_taxon["_source"]["lineage"] = lineage
    return desc_taxon


def add_new_taxon(alt_taxon_id, new_taxa, obj, closest_taxon, *, blanks={"NA", "None"}):
    """Add a new taxon with alt_taxon_id to list of taxa."""
    # TODO: Allow creation of new higher taxa
    if alt_taxon_id not in new_taxa:
        alt_rank = "species"
        if (
            "subspecies" in obj["taxonomy"]
            and obj["taxonomy"]["subspecies"]
            and obj["taxonomy"]["subspecies"] not in blanks
        ):
            alt_rank = "subspecies"
        new_taxon = create_descendant_taxon(
            alt_taxon_id, alt_rank, obj["taxonomy"][alt_rank], closest_taxon
        )
        new_taxa.update({new_taxon["_source"]["taxon_id"]: new_taxon["_source"]})
    return new_taxon


def create_taxa(es, opts, *, taxon_template, data=None, blanks=set(["NA", "None"])):
    """Create new taxa using alternate taxon IDs."""
    default_ranks = [
        "genus",
        "family",
        "order",
        "class",
        "subphylum",
        "phylum",
    ]
    ancestors = {}
    matches = defaultdict(dict)
    pbar = tqdm(total=len(data.keys()))
    taxon_ids = set({})
    new_taxa = {}
    for alt_taxon_id, rows in data.items():
        obj = rows[0]
        pbar.update(1)
        if "taxonomy" not in obj:
            continue
        lineage = []
        closest_rank = None
        closest_taxon = None
        if "subspecies" in obj["taxonomy"]:
            ranks = ["species"] + default_ranks
        else:
            ranks = default_ranks
        max_index = len(ranks) - 1
        # max_rank = ranks[max_index]
        for index, rank in enumerate(ranks[: (max_index - 1)]):
            if rank not in obj["taxonomy"] or obj["taxonomy"][rank] in blanks:
                continue
            intermediates = 0
            for anc_rank in ranks[(index + 1) :]:
                if (
                    anc_rank not in obj["taxonomy"]
                    or obj["taxonomy"][anc_rank] in blanks
                ):
                    continue
                if (
                    obj["taxonomy"][rank] in matches
                    and obj["taxonomy"][anc_rank] in matches[obj["taxonomy"][rank]]
                ):
                    taxa = matches[obj["taxonomy"][rank]][obj["taxonomy"][anc_rank]]
                    ancestors.update({alt_taxon_id: taxa[0]})
                    break
                else:
                    taxa = lookup_taxon_within_lineage(
                        es,
                        obj["taxonomy"][rank],
                        obj["taxonomy"][anc_rank],
                        opts,
                        rank=rank,
                        anc_rank=anc_rank,
                        return_type="taxon",
                    )
                if taxa:
                    if len(taxa) == 1:
                        ancestors.update({alt_taxon_id: taxa[0]})
                        matches[obj["taxonomy"][rank]][obj["taxonomy"][anc_rank]] = taxa
                        break
                # elif anc_rank == max_rank and intermediates == 0:
                elif intermediates == 0:
                    taxa, name_class = lookup_taxon(
                        es,
                        obj["taxonomy"][anc_rank],
                        opts,
                        rank=anc_rank,
                        return_type="taxon",
                    )
                    if taxa and len(taxa) == 1:
                        ancestors.update({alt_taxon_id: taxa[0]})
                        matches[obj["taxonomy"][anc_rank]]["all"] = taxa
                        break
                intermediates += 1
            if alt_taxon_id in ancestors:
                closest_rank = rank
                if (
                    obj["taxonomy"][rank] in matches
                    and obj["taxonomy"][anc_rank] in matches[obj["taxonomy"][rank]]
                ):
                    closest_taxon = matches[obj["taxonomy"][rank]][
                        obj["taxonomy"][anc_rank]
                    ][0]
                else:
                    closest_taxon = matches[obj["taxonomy"][anc_rank]]["all"][0]
                break
            lineage.append({"rank": rank, "name": obj["taxonomy"][rank]})
        if closest_taxon is not None:
            for intermediate in reversed(lineage):
                taxon_id = generate_ancestral_taxon_id(
                    intermediate["name"],
                    intermediate["rank"],
                    alt_taxon_id=alt_taxon_id,
                    taxon_ids=taxon_ids,
                )
                new_taxon = create_descendant_taxon(
                    taxon_id, intermediate["rank"], intermediate["name"], closest_taxon
                )
                new_taxa.update(
                    {new_taxon["_source"]["taxon_id"]: new_taxon["_source"]}
                )
                matches[intermediate["name"]][obj["taxonomy"][closest_rank]] = taxa
                closest_rank = intermediate["rank"]
                closest_taxon = new_taxon
            ancestors[alt_taxon_id] = closest_taxon
            added_taxon = add_new_taxon(alt_taxon_id, new_taxa, obj, closest_taxon)
            matches[added_taxon["_source"]["scientific_name"]][
                closest_taxon["_source"]["scientific_name"]
            ] = [added_taxon]
    pbar.close()
    index_stream(
        es, taxon_template["index_name"], stream_taxa(new_taxa),
    )
    return new_taxa.keys()


# def parse_taxa(es, types, taxonomy_template):
#     """Test method to parse taxa."""
#     taxa = [
#         {
#             "taxon_id": 110368,
#             "assembly_span": 12344567,
#             "c_value": 2.5,
#             "sex_determination_system": "N/A",
#         },
#         {
#             "taxon_id": 13037,
#             "assembly_span": 2345678,
#             "c_value": 2.3,
#             "sex_determination_system": "XO",
#         },
#         {
#             "taxon_id": 113334,
#             "assembly_span": 45678912,
#             "c_value": 4.6,
#             "sex_determination_system": "XY",
#         },
#     ]
#     for entry in taxa:
#         # attributes = {}
#         taxon_id = str(entry["taxon_id"])
#         doc = lookup_taxon_by_taxid(es, taxon_id, taxonomy_template)
#         if doc is None:
#             LOGGER.warning(
#                 "No %s taxonomy record for %s",
#                 taxonomy_template["index_name"],
#                 taxon_id,
#             )
#         attributes = add_attributes(entry, types, attributes=[])[0]
#         doc.update({"taxon_id": taxon_id, "attributes": attributes})
#         doc_id = "taxon_id-%s" % taxon_id
#         yield doc_id, doc


# def index(es, opts, *, taxonomy_name="ncbi"):
#     """Index a set of taxa."""
#     LOGGER.info("Indexing taxa using %s taxonomy", taxonomy_name)
#     template = index_template(taxonomy_name, opts)
#     taxonomy_template = taxonomy_index_template(taxonomy_name, opts)
#     stream = parse_taxa(es, template["types"], taxonomy_template)
#     return template, stream



#!/usr/bin/env python3

"""
Initialise a GenomeHubs instance.

Usage:
    genomehubs init [--hub-name STRING] [--hub-path PATH] [--hub-version PATH]
                    [--config-file PATH...] [--config-save PATH]
                    [--es-host URL...] [--es-url URL]
                    [--insdc-metadata] [--insdc-root INT...] [--restore-indices]
                    [--taxonomy-path PATH] [--taxonomy-ncbi-root INT]
                    [--taxonomy-ncbi-url URL] [--taxon-preload]
                    [--docker-contain STRING...] [--docker-network STRING]
                    [--docker-timeout INT] [--docker-es-container STRING]
                    [--docker-es-image URL]
                    [--reset] [--force-reset]
                    [-h|--help] [-v|--version]

Options:
    --hub-name STRING             GenomeHubs instance name.
    --hub-path PATH               GenomeHubs instance root directory path.
    --hub-version STR             GenomeHubs instance version string.
    --config-file PATH            Path to YAML file containing configuration options.
    --config-save PATH            Path to write configuration options to YAML file.
    --es-host URL                 ElasticSearch hostname/URL and port.
    --es-url URL                  Remote URL to fetch ElasticSearch code.
    --insdc-metadata              Flag to index metadata for public INSDC assemblies.
    --insdc-root INT              Root taxid when indexing public INSDC assemblies.
    --restore-indices             Flag to restore taxon and assembly indices.
    --taxonomy-path DIR           Path to directory containing raw taxonomies.
    --taxonomy-ncbi-root INT      Root taxid for NCBI taxonomy index.
    --taxonomy-ncbi-url URL       Remote URL to fetch NCBI taxonomy.
    --taxon-preload               Flag to preload all taxa in taxonomy into taxon index.
    --docker-contain STRING       GenomeHubs component to run in Docker.
    --docker-network STRING       Docker network name.
    --docker-timeout STRING       Time in seconds to wait for a component to start in
                                  Docker.
    --docker-es-container STRING  ElasticSearch Docker container name.
    --docker-es-image STRING      ElasticSearch Docker image name.
    --reset                       Flag to reset GenomeHubs instance if already exists.
    --force-reset                 Flag to force reset GenomeHubs instance if already
                                  exists.
    -h, --help                    Show this
    -v, --version                 Show version number

Examples:
    # 1. New GenomeHub with default settings
    ./genomehubs init

    # 2. New GenomeHub in specified directory, populated with Lepidoptera assembly
    #    metadata from INSDC
    ./genomehubs init --hub-path /path/to/GenomeHub --insdc-root 7088 --insdc-meta
"""

import sys

from docopt import docopt
from tolkein import tolog

from ..lib import analysis
from ..lib import assembly
from ..lib import es_functions
from ..lib import files
from ..lib import hub
from ..lib import taxon
from ..lib import taxonomy
from .config import config
from .version import __version__

LOGGER = tolog.logger(__name__)


def main(args):
    """Initialise genomehubs."""
    options = config("init", **args)

    # setup GenomeHubs directory
    hub.setup(options["init"])

    # Start Elasticsearch
    es = es_functions.launch_es(options["init"])

    # Post search scripts
    hub.post_search_scripts(es)

    # Index taxonomies
    if "taxonomy-source" in options["init"]:
        for taxonomy_name in options["init"]["taxonomy-source"]:
            template, stream = taxonomy.index(taxonomy_name, options["init"])
            if "taxonomy-%s-root" % taxonomy_name in options["init"]:
                es_functions.load_mapping(es, template["name"], template["mapping"])
                es_functions.index_stream(es, template["index_name"], stream)

            # Prepare taxon index
            taxon_template = taxon.index_template(taxonomy_name, options["init"])
            es_functions.load_mapping(
                es, taxon_template["name"], taxon_template["mapping"]
            )
            # attributes.index_types(
            #     es, taxon_template["name"], taxon_template["types"], options["init"]
            # )
            es_functions.index_create(es, taxon_template["index_name"])
            if options["init"].get("taxon-preload", False):
                LOGGER.info("Loading all taxa from taxonomy into taxon index")
                body = {
                    "source": {"index": template["index_name"]},
                    "dest": {"index": taxon_template["index_name"]},
                }
                es.reindex(body=body)

            # Prepare assembly index
            assembly_template = assembly.index_template(taxonomy_name, options["init"])
            es_functions.load_mapping(
                es, assembly_template["name"], assembly_template["mapping"]
            )
            es_functions.index_create(es, assembly_template["index_name"])
            # attributes.index_types(
            #     es, assembly_template["name"], assembly_template["types"], options["init"]
            # )

            # if "insdc-metadata" in options["init"]:
            #     assembly_metadata.index(
            #         es, options["init"], metadata_name="insdc", taxonomy_name=taxonomy_name
            #     )
            # Prepare analysis index
            analysis_template = analysis.index_template(taxonomy_name, options["init"])
            es_functions.load_mapping(
                es, analysis_template["name"], analysis_template["mapping"]
            )
            es_functions.index_create(es, analysis_template["index_name"])

            # Prepare file index
            file_template = files.index_template(taxonomy_name, options["init"])
            es_functions.load_mapping(
                es, file_template["name"], file_template["mapping"]
            )
            es_functions.index_create(es, file_template["index_name"])


def cli():
    """Entry point."""
    if len(sys.argv) == sys.argv.index(__name__.split(".")[-1]) + 1:
        args = docopt(__doc__, argv=[])
    else:
        args = docopt(__doc__, version=__version__)
    main(args)


if __name__ == "__main__":
    cli()



#!/usr/bin/env python3

"""Taxon methods."""

from tolkein import tolog

from .es_functions import index_stream
from .es_functions import load_mapping
from .es_functions import stream_template_search_results
from .hub import index_templator

LOGGER = tolog.logger(__name__)


def index_template(opts):
    """Index template (includes name, mapping and types)."""
    parts = ["attributes", opts["hub-name"], opts["hub-version"]]
    template = index_templator(parts, opts)
    return template


def stream_attributes(group, attributes):
    """Stream attributes for indexing."""
    for name, obj in attributes.items():
        ret = {"group": group, "name": name}
        for prop, value in obj.items():
            if not prop.startswith("taxon_"):
                ret.update({prop: value})
        yield "attribute-%s-%s" % (group, name), ret


def index(es, group, attributes, opts):
    """Index a set of attributes."""
    LOGGER.info("Indexing attributes")
    template = index_template(opts)
    stream = stream_attributes(group, attributes)
    return template, stream


def index_types(es, types_name, types, opts):
    """Index types into Elasticsearch."""
    if "attributes" not in types:
        return
    if "defaults" in types:
        for key, value in types["attributes"].items():
            value = {**types["defaults"]["attributes"], **value}
            types["attributes"][key] = value
    template, stream = index(es, types_name, types["attributes"], opts)
    load_mapping(es, template["name"], template["mapping"])
    index_stream(es, template["index_name"], stream)


def fetch_types(es, types_name, opts):
    """Fetch types from Elasticsearch."""
    template = index_template(opts)
    body = {"id": "attribute_types_by_group", "params": {"group": types_name}}
    types = {}
    for result in stream_template_search_results(
        es, index=template["index_name"], body=body, size=50
    ):
        source = result["_source"]
        name = source["name"]
        del source["name"]
        del source["group"]
        types[name] = source
    return types



#!/usr/bin/env python3

"""
Index a file, directory or repository.

Usage:
    genomehubs index [--hub-name STRING] [--hub-path PATH] [--hub-version PATH]
                     [--config-file PATH...] [--config-save PATH]
                     [--es-host URL...] [--assembly-dir PATH]
                     [--assembly-repo URL] [--assembly-exception PATH]
                     [--taxon-dir PATH] [--taxon-repo URL] [--taxon-exception PATH]
                     [--taxon-lookup STRING] [--file PATH...] [file-dir PATH...]
                     [--remote-file URL...] [--remote-file-dir URL...]
                     [--taxon-id STRING] [--assembly-id STRING] [--analysis-id STRING]
                     [--file-title STRING] [--file-description STRING] [--file-metadata PATH]
                     [-h|--help] [-v|--version]

Options:
    --hub-name STRING          GenomeHubs instance name.
    --hub-path PATH            GenomeHubs instance root directory path.
    --hub-version STR          GenomeHubs instance version string.
    --config-file PATH         Path to YAML file containing configuration options.
    --config-save PATH         Path to write configuration options to YAML file.
    --es-host URL              ElasticSearch hostname/URL and port.
    --assembly-dir PATH        Path to directory containing assembly-level data.
    --assembly-repo URL        Remote git repository containing assembly-level data.
                               Optionally include `~branch-name` suffix.
    --assembly-exception PATH  Path to directory to write assembly data that failed to import.
    --taxon-lookup STRING      Taxon name class to lookup (scientific|all). [Default: scientific]
    --taxon-dir PATH           Path to directory containing taxon-level data.
    --taxon-repo URL           Remote git repository containing taxon-level data.
                               Optionally include `~branch-name` suffix.
    --taxon-exception PATH     Path to directory to write taxon data that failed to import.
    --file PATH                Path to file for generic file import.
    --file-dir PATH            Path to directory containing generic files to import.
    --remote-file URL          Location of remote file for generic file import.
    --remote-file-dir URL      Location of remote directory containing generic files to import.
    --taxon-id STRING          Taxon ID to index files against.
    --assembly-id STRING       Assembly ID to index files against.
    --analysis-id STRING       Analysis ID to index files against.
    --file-title STRING        Default title for indexed files.
    --file-description STRING  Default description for all indexed files.
    --file-metadata PATH       CSV, TSV, YAML or JSON file metadata with one entry per file to be indexed.
    -h, --help                 Show this
    -v, --version              Show version number

Examples:
    # 1. Index all files in a remote repository
    ./genomehubs index --taxon-repo https://github.com/example/repo~main
"""

import csv
import sys
from collections import defaultdict
from pathlib import Path

from docopt import docopt
from tolkein import tolog
from tqdm import tqdm

from ..lib import assembly
from ..lib import es_functions
from ..lib import hub
from ..lib import taxon
from .assembly import add_identifiers_and_attributes_to_assemblies
from .attributes import index_types
from .config import config
from .es_functions import index_stream
from .files import index_files
from .files import index_metadata
from .hub import process_row
from .hub import set_column_indices
from .hub import validate_types_file
from .taxon import add_names_and_attributes_to_taxa
from .taxon import fix_missing_ids
from .version import __version__

LOGGER = tolog.logger(__name__)


def index_file(es, types, data, opts):
    """Index a file."""
    delimiters = {"csv": ",", "tsv": "\t"}
    rows = csv.reader(
        data, delimiter=delimiters[types["file"]["format"]], quotechar='"'
    )
    header = None
    if types["file"].get("header", False):
        header = next(rows)
        set_column_indices(types, header)
    with_ids = defaultdict(list)
    taxon_asm_data = defaultdict(list)
    without_ids = defaultdict(list)
    failed_rows = defaultdict(list)
    blanks = set(["", "NA", "N/A", "None"])
    taxon_types = {}
    for taxonomy_name in opts["taxonomy-source"]:
        taxon_template = taxon.index_template(taxonomy_name, opts)
        LOGGER.info("Processing rows")
        for row in tqdm(rows):
            try:
                processed_data, taxon_data, new_taxon_types = process_row(types, row)
            except Exception:
                failed_rows["None"].append(row)
                continue
            taxon_types.update(new_taxon_types)
            if (
                "taxon_id" in processed_data["taxonomy"]
                and processed_data["taxonomy"]["taxon_id"] not in blanks
            ):
                with_ids[processed_data["taxonomy"]["taxon_id"]].append(processed_data)
                taxon_asm_data[processed_data["taxonomy"]["taxon_id"]].append(
                    taxon_data
                )
            else:
                if "taxonomy" in types and "alt_taxon_id" in types["taxonomy"]:
                    without_ids[processed_data["taxonomy"]["alt_taxon_id"]].append(
                        processed_data
                    )
                    taxon_asm_data[processed_data["taxonomy"]["alt_taxon_id"]].append(
                        taxon_data
                    )
                    failed_rows[processed_data["taxonomy"]["alt_taxon_id"]].append(row)
                elif "subspecies" in processed_data["taxonomy"]:
                    without_ids[processed_data["taxonomy"]["subspecies"]].append(
                        processed_data
                    )
                    taxon_asm_data[processed_data["taxonomy"]["subspecies"]].append(
                        taxon_data
                    )
                    failed_rows[processed_data["taxonomy"]["subspecies"]].append(row)
                elif "species" in processed_data["taxonomy"]:
                    without_ids[processed_data["taxonomy"]["species"]].append(
                        processed_data
                    )
                    taxon_asm_data[processed_data["taxonomy"]["species"]].append(
                        taxon_data
                    )
                    failed_rows[processed_data["taxonomy"]["species"]].append(row)
                else:
                    failed_rows["None"].append(row)
        LOGGER.info("Found taxon IDs in %d entries", len(with_ids.keys()))
        create_ids, without_ids = fix_missing_ids(
            es,
            opts,
            without_ids,
            types=types,
            taxon_template=taxon_template,
            failed_rows=failed_rows,
            with_ids=with_ids,
            blanks=blanks,
            header=header,
        )
        if with_ids or create_ids:
            LOGGER.info("Indexing %d entries", len(with_ids.keys()))
            if opts["index"] == "taxon":
                docs = add_names_and_attributes_to_taxa(
                    es, dict(with_ids), opts, template=taxon_template, blanks=blanks
                )
                index_stream(
                    es, taxon_template["index_name"], docs, _op_type="update",
                )
            elif opts["index"] == "assembly":
                assembly_template = assembly.index_template(taxonomy_name, opts)
                docs = add_identifiers_and_attributes_to_assemblies(
                    es,
                    with_ids,
                    opts,
                    template=assembly_template,
                    taxon_template=taxon_template,
                    blanks=blanks,
                )
                index_stream(es, assembly_template["index_name"], docs)
                # index taxon-level attributes
                index_types(
                    es, "taxon", {"attributes": taxon_types}, opts,
                )
                taxon_asm_with_ids = {
                    taxon_id: taxon_asm_data[taxon_id] for taxon_id in with_ids.keys()
                }
                taxon_docs = add_names_and_attributes_to_taxa(
                    es, taxon_asm_with_ids, opts, template=taxon_template, blanks=blanks
                )
                index_stream(
                    es, taxon_template["index_name"], taxon_docs, _op_type="update",
                )


def main(args):
    """Index files."""
    options = config("index", **args)

    # Start Elasticsearch
    es = es_functions.launch_es(options["fill"])

    # Post search scripts
    with tolog.DisableLogger():
        hub.post_search_scripts(es)

    for index in list(["taxon", "assembly"]):
        data_dir = "%s-dir" % index
        if data_dir in options["index"]:
            dir_path = options["index"][data_dir]
            for types_file in sorted(Path(dir_path).glob("*.names.yaml")):
                types, data = validate_types_file(types_file, dir_path)
                LOGGER.info("Indexing %s" % types["file"]["name"])
                index_types(es, index, types, options["index"])
                index_file(
                    es,
                    types,
                    data,
                    {**options["index"], "index": index, "index_types": index_types},
                )
            for types_file in sorted(Path(dir_path).glob("*.types.yaml")):
                types, data = validate_types_file(types_file, dir_path)
                LOGGER.info("Indexing %s" % types["file"]["name"])
                index_types(es, index, types, options["index"])
                index_file(
                    es,
                    types,
                    data,
                    {**options["index"], "index": index, "index_types": index_types},
                )
    # TODO: #29 Implement alternate backbone taxonomies
    taxonomy_name = options["index"]["taxonomy-source"][0]
    if "file" in options["index"]:
        index_files(es, options["index"]["file"], taxonomy_name, options["index"])
    elif "file-metadata" in options["index"]:
        index_metadata(
            es, options["index"]["file-metadata"], taxonomy_name, options["index"]
        )


def cli():
    """Entry point."""
    if len(sys.argv) == sys.argv.index("index") + 1:
        args = docopt(__doc__, argv=[])
    else:
        args = docopt(__doc__, version=__version__)
    main(args)


if __name__ == "__main__":
    cli()
